## destine_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import xarray as xr
from polytope.api import Client
import glob
import shutil
from cdo import Cdo
from importlib.resources import files
from .dictionary import dictionary


def load_destine(model_global, variable, experiment, start, end,
                 output_path):
    """
    Download hourly Destination Earth DT data via the Polytope web service
    hosted on the LUMI Databridge.

        Input:
            model_global    str     name of the global model
            variable        str     variable name (e.g. 2 metre temperature)
            experiment      str     climate change scenario
            start           str     start (e.g. 20200901)
            end             str     end (e.g. 20200930)
            output_path     str     path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the DestinE
            daily mean data calculated from downloaded hourly data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    destine_info = dictionary[variable]['destine']
    variable_long = destine_info['name']
    standard_key = next(iter(dictionary[variable]))
    standard_unit = dictionary[variable][standard_key]['units']

    # Define filename
    file = f"{variable}_{model_global}_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded DestinE Climate DT data successfully.")
        return output_file

    else:
        # Create temporary directory
        temp_dir = os.path.join(output_path, "tmp")
        os.makedirs(temp_dir, exist_ok=True)
        temp_file = os.path.join(temp_dir, file)

        # Define range of dates to be downloaded
        dates = xr.cftime_range(
            start=start,
            end=end,
            freq='1D',
            inclusive='both',
            calendar='proleptic_gregorian'
        )

        for date in dates:
            date = date.strftime('%Y%m%d')

            fn = os.path.join(temp_dir, f'{model_global}-climate-dt_{variable}_high_hourly_{date}.grib')
            if os.path.isfile(fn):
                continue

            # Initialize client
            client = Client(
                address="polytope.lumi.apps.dte.destination-earth.eu",
            )

            # Optionally revoke previous requests
            client.revoke("all")

            # Check experiment/activity consistency for scenario simulation
            if experiment == 'SSP3-7.0':
                activity = 'ScenarioMIP'

            # Check experiment/activity consistency for historical simulation
            if experiment == 'hist':
                activity = 'CMIP6'

            request = {
                "activity": activity,
                "class": "d1",
                "dataset": "climate-dt",
                "date": date,
                "experiment": experiment,
                "expver": "0001",
                "generation": "1",
                "levtype": "sfc",
                "model": model_global,
                "param": variable_long,
                "realization": "1",
                "resolution": "high",
                "stream": "clte",
                "time": "0000/0100/0200/0300/0400/0500/0600/0700/0800/0900/1000/1100/1200/1300/1400/1500/1600/1700/1800/1900/2000/2100/2200/2300",
                "type": "fc",
            }

            try:
                retrieve = client.retrieve(
                    "destination-earth",
                    request,
                    output_file=fn,
                    pointer=False)
                print(f'{date=}...DONE')

            except Exception:
                print(f'{date=}...FAIL')
                continue

        # Glob all GRIB files in the directory
        grib_files = sorted(glob.glob(os.path.join(temp_dir, "*.grib")))

        for grib_file in grib_files:
            # Open the GRIB file
            dataset = xr.open_dataset(grib_file, engine="cfgrib")

            # Standardize variable and units for temperature
            if variable == 'tas':
                # Rename the variable
                dataset = dataset.rename({'t2m': variable})

                # Calculate the daily mean
                daily_mean = dataset.resample(valid_time="1D").mean(dim="time")

                # Rename the time coordinate
                daily_mean = daily_mean.rename({'valid_time': 'time'})

                # Convert from Kelvin to Celsius
                daily_mean[variable] = daily_mean[variable] - 273.15
                units = daily_mean[variable].attrs.get("units", None)
                if units is None or units not in standard_unit:
                    print(f"Warning: {variable} has no or wrong unit attribute.")
                    daily_mean[variable].attrs['units'] = standard_unit
                    print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

                # Generate an output file name
                base_name = os.path.basename(grib_file).replace(".grib", "_daily_mean.nc")
                daily_file = os.path.join(temp_dir, base_name)

                # Save the daily mean to a NetCDF file
                daily_mean.to_netcdf(daily_file)

            if variable == 'pr':
                # Rename the variable
                dataset = dataset.rename({'tprate': variable})

                # Calculate the daily mean
                daily_mean = dataset.resample(valid_time="1D").mean(dim="time")

                # Rename the time coordinate
                daily_mean = daily_mean.rename({'valid_time': 'time'})

                # Convert from kg/m² second into kg/m² day
                daily_mean[variable] = daily_mean[variable]*86400

                # Generate an output file name
                base_name = os.path.basename(grib_file).replace(".grib", "_daily_mean.nc")
                daily_file = os.path.join(temp_dir, base_name)

                # Save the daily mean to a NetCDF file
                daily_mean.to_netcdf(daily_file)

        # Use CDO to merge netcdf files
        daily_files = sorted(glob.glob(os.path.join(temp_dir, "*_daily_mean.nc")))

        if daily_files:
            # Convert the list of files into a space-separated string
            file_list = " ".join(daily_files)

            # If cdo is not in the path, add it manually
            conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
            os.environ['PATH'] += f':{conda_bin}'

            # Initialize the Cdo object
            cdo = Cdo()

            # Perform merging of files on the time dimension
            cdo.mergetime(input=file_list, output=temp_file)

            # Dynamically get the file path to healpix_grid.txt inside the package
            healpix_grid_path = files('climxtract').joinpath('healpix_grid.txt')

            # Set grid from unstructured (default Destination Earth) to healpix for conservative regridding
            cdo.setgrid(str(healpix_grid_path), input=temp_file, output=output_file)

        # Clean temporary directory
        shutil.rmtree(temp_dir)

        return output_file



# cut.py
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import numpy as np
import xarray as xr
import netCDF4
import os


def masking(target_grid, source_grid, output_path_mask):
    """
    Masking the data to match the spatial extent of target data.

        Input:
            target_grid              str        path to target grid
            source_grid              str        path to source grid
            output_path_mask         str        base path where the output file is stored

        Output:
            Returns mask netcdf-file based on nan-values of the target grid.
    """
    # Extract the filename from the source_grid
    filename = source_grid.split('/')[-1]

    # Check if the filename contains one of the specified substrings
    remap_options = ["_remapnn.nc", "_remapbil.nc", "_remapcon.nc", "_remapdis.nc"]

    # Modify the filename to include '_mask' before the .nc extension
    for remap_option in remap_options:
        if filename.endswith(remap_option):
            file = filename.replace(".nc", "_mask.nc")

    output_file = f"{output_path_mask}/{file}"

    # check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        return output_file
    else:
        # read in target and source grid
        target_grid = xr.open_dataset(target_grid)
        source_grid = xr.open_dataset(source_grid)

        if 'tas' in source_grid:
            variable = 'tas'

        if 'pr' in source_grid:
            variable = 'pr'

        nan_locations = target_grid[variable][0].isnull()
        source_mask = source_grid[variable].where(~nan_locations, np.nan)

        # save xarray Dataset to netcdf-file
        source_mask.to_netcdf(output_file)
        return output_file


# spartacus_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import shutil
import wget
import xarray as xr
from cdo import Cdo
from importlib.resources import files
from .dictionary import dictionary


def download_yearly_files(dataset, year, output_path):
    """
    Download daily SPARTACUS v2.1 data on a yearly basis
    from the Geosphere Austria Datahub.

        Input:
            dataset    str      type of the dataset (e.g tn)
            year       int      year to be downloaded (e.g 1971)

        Output:
            Returns the temporary output path, where the yearly downloaded
            files are saved.
    """
    # Putting together the download link (url)
    base_url = (
        "https://public.hub.geosphere.at/datahub/resources/"
        "spartacus-v2-1d-1km/filelisting"
    )

    url_directory = f"{base_url}/{dataset.upper()}/"
    file = f"SPARTACUS2-DAILY_{dataset.upper()}_{year}.nc"

    # Construct the full URL
    url = f"{url_directory}{file}"

    # Define temporary directory
    temp_dir = os.path.join(output_path, "tmp")

    # Ensure the temporary directory exsists
    os.makedirs(temp_dir, exist_ok=True)

    # Define filname for temporal output
    temp_file = os.path.join(temp_dir, file)

    # Download the file using the wget package
    wget.download(url, out=temp_file)

    return temp_file


def load_spartacus(variable, start, end, output_path):
    """
    Function that calls download_yearly_files to download SPARTACUS data.
    The individual downloaded years are merged into one file.
    The temperature is calculated as the mean of daily maximum and
    minimum temperature.

        Input:
            variable          str      variable name (e.g. tas)
            start             str      start year (e.g. 1961)
            end               str      end year (e.g. 2024)
            output_path       str      path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the SPARTACUS data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    spartacus_info = dictionary[variable]['spartacus']
    variable_long = spartacus_info['name']
    standard_key = next(iter(dictionary[variable]))
    standard_unit = dictionary[variable][standard_key]['units']

    # Define filename
    tmp_file = (
        f"{variable}_spartacusv2.1_"
        f"{start}-{end}_tmp.nc"
    )

    file = (
        f"{variable}_spartacusv2.1_"
        f"{start}-{end}.nc"
    )

    tmp_file = os.path.join(output_path, tmp_file)
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded SPARTACUS data successfully.")
        return output_file

    else:

        try:
            # Initialize list to hold data arrays
            data_list = []

            # Loop through years and download data
            for year in range(int(start), int(end) + 1):
                if variable == 'tas':
                    # Download tn and tx datasets
                    tasmin_path = download_yearly_files('tn', year, output_path)
                    tasmax_path = download_yearly_files('tx', year, output_path)

                    # Load the datasets
                    tasmin = xr.open_dataset(tasmin_path)
                    tasmax = xr.open_dataset(tasmax_path)

                    # Append the data arrays to the list
                    data_list.append((tasmin['TN'] + tasmax['TX']) / 2)

                elif variable == 'pr':
                    # Download rr dataset
                    rr_path = download_yearly_files('rr', year, output_path)

                    # Load the dataset
                    rr = xr.open_dataset(rr_path)

                    # Appaned the data arrays to the list
                    data_list.append(rr['RR'])

            # Concatenate all years together along the time dimension
            dataset = xr.concat(data_list, dim='time')

            # Standardize variable and units for temperature
            if variable == 'tas':
                dataset.name = variable
                units = dataset.attrs.get("units", None)
                if units is None or units not in standard_unit:
                    dataset.attrs['units'] = standard_unit

            # Standardize varibale and units for precipitation
            if variable == 'pr':
                dataset.allname = variable
                units = dataset.attrs.get("units", None)
                if units is None or units not in standard_unit:
                    dataset.attrs['units'] = standard_unit

            # Save dataset to netCDF file
            dataset.to_netcdf(tmp_file)
            print("Downloaded SPARTACUS data successfully.")

            # If cdo is not in the path, add it manually
            conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
            os.environ['PATH'] += f':{conda_bin}'

            # Initialize the Cdo object
            cdo = Cdo()

            # Dynamically get the file path to spartacus_grid.txt inside the package
            spartacus_grid_path = files('climxtract').joinpath('spartacus_grid.txt')

            # Set spartacus grid for conservative regridding
            cdo.setgrid(str(spartacus_grid_path), input=tmp_file, output=output_file)

        except Exception as e:
            print(f"Failed to download SPARTACUS data.\nError: {e}")
            return None

        # Clean temporary directory
        shutil.rmtree(os.path.join(output_path, "tmp"))

        # Remove tmp_file with old grid
        os.remove(tmp_file)

        return output_file



# dictionary.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

# Dictionary for datasets, variables and units
dictionary = {
    'tas': {
        'oeks15': {'name': 'temperature', 'units': ['Celsius']},
        'spartacus': {'name': 'temperature', 'units': ['Celsius']},
        'eurocordex': {'name': 'tas', 'units': ['Kelvin']},
        'eobs': {'name': 'tg', 'units': ['Celsius']},
        'destine': {'name': '167', 'units': ['Kelvin']},
        'era5': {'name': '2m_temperature', 'units': ['Kelvin']},
        'ch2025': {'name': 'air_temperature', 'units': ['degrees_C']},
    },
    'pr': {
        'oeks15': {'name': 'precipitation', 'units': ['kg m-2']},
        'spartacus': {'name': 'RR', 'units': ['kg m-2']},
        'eurocordex': {'name': 'pr', 'units': ['kg m-2 s-1']},
        'eobs': {'name': 'rr', 'units': ['mm/day', 'mm d-1']},
        'destine': {'name': '260048', 'units': ['kg m-2 s-1']},
        'era5': {'name': 'total_precipitation', 'units': ['m']},
        'ch2025': {'name': 'precipitation_sum', 'units': ['mm day-1']}
    }
}


# oeks15_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import wget
import os
from .dictionary import dictionary


def load_oeks15(model_global, model_regional, variable,
                experiment, ens, output_path):
    """
    Download ÖKS15 data from the Geosphere Austria Datahub and save as netCDF
    file in the corresponding output path. The complete dataset from 1951-2100
    is downloaded.

        Input:
            model_global         string      name of the global climate model
            model_regional       string      name of the regional model
            variable             string      variable name (e.g. tas)
            experiment           string      climate change scenario
            ens                  string      name of ensemble member
            output_path          string      path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the ÖKS15 data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    oeks15_info = dictionary[variable]['oeks15']
    variable_long = oeks15_info['name']
    #standard_key = next(iter(dictionary[variable]))
    #standard_unit = dictionary[variable][standard_key]['units']

    # Define filename
    file = (
        f"{variable}_SDM_{model_global}_{experiment}_"
        f"{ens}_{model_regional}.nc"
    )

    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded ÖKS15 data successfully.")
        return output_file

    else:
        # Check version for temperature and maximum temperature
        if variable == 'tas' or variable == 'tx':
            version = '-v02/'

        # Check version for precipitation
        if variable == 'pr':
            version = '-v01/'

        # Download ÖKS15 data
        try:
            # Put together the download link (url)
            base_url = (
                "https://public.hub.geosphere.at/public/resources/oks15/"
                "bias_corrected/oks15_bias_corrected_"
            )

            url_directory = (
                f"{base_url}{variable_long}_{variable}_"
                f"{str.lower(model_global)}_{experiment}_"
                f"{ens}_{str.lower(model_regional)}{version}"
            )

            # Construct the full URL
            url = f"{url_directory}{file}"

            # Download the file using wget
            output_file = wget.download(url, out=output_file)
            print("Downloaded ÖKS15 data successfully.")

        except Exception as e:
            print(f"Failed to download ÖKS15 data.\nError: {e}")
            return None

        return output_file



# eobs_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import wget
import shutil
import xarray as xr
from .dictionary import dictionary


def load_eobs(variable, start, end, output_path):
    """
    Download daily E-OBS data from European Climate Assessment & Dataset.
    The complete dataset from 1950 onwards is downloaded.

        Input:
            variable       str          variable name (e.g. 2m_temperature)
            start          str          start date (20200101)
            end            str          end date (20201231)
            output_path    str          path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the daily E-OBS data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    eobs_info = dictionary[variable]['eobs']
    variable_long = eobs_info['name']
    standard_key = next(iter(dictionary[variable]))
    standard_unit = dictionary[variable][standard_key]['units']

    # Define output filename
    file = f"{variable}_e-obs31.e_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded E-OBS data successfully.")
        return output_file

    else:
        # Create temporary directory
        temp_dir = os.path.join(output_path, "tmp")
        os.makedirs(temp_dir, exist_ok=True)
        temp_file = os.path.join(temp_dir, file)

        # Download E-OBS data
        try:
            base_url = "https://knmi-ecad-assets-prd.s3.amazonaws.com/ensembles/data/"
            url_directory = base_url + "Grid_0.1deg_reg_ensemble/"
            filename = str(variable_long) + "_" + "ens_mean_0.1deg_reg_v30.0e.nc"
            url = url_directory + filename
            wget.download(url, out=temp_file)
            print("Downloaded E-OBS data successfully.")
        except Exception as e:
            print(f"Failed to download E-OBS data.\nError: {e}")
            return None

        # Open previously downloaded dataset
        dataset = xr.open_dataset(temp_file)

        # Select the data range
        dataset = dataset.sel(time=slice(start, end))

        # Standardize variable and units for temperature
        if variable == 'tas':
            dataset = dataset.rename({variable_long: variable})
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

        # Standardize variable and units for precipitation
        if variable == 'pr':
            dataset = dataset.rename({variable_long: variable})
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

        # Write dataset to netCDF file
        dataset.to_netcdf(output_file)

        # Clean temporary directory
        shutil.rmtree(temp_dir)

        return output_file



# __init__.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import xarray as xr
from .oeks15_download import load_oeks15
from .spartacus_download import load_spartacus
from .cordex_download import load_cordex
from .eobs_download import load_eobs
from .destine_download import load_destine
from .era5_download import load_era5
from .ch2025_download import load_ch2025
from .regrid import remapbil
from .regrid import remapnn
from .regrid import remapcon
from .regrid import remapdis
from .mask import masking


def load(type, model_global, model_regional, variable, experiment, ens, start, end, output_path):
    """
    Function that calls the corresponding load functions, depending on the chosen type.
    Both the path of the file and the xarray object are returned.
    """
    if type == 'oeks15':
        # Call the load_oeks15 function to get the file path
        file_path_oeks15 = load_oeks15(model_global, model_regional, variable, experiment, ens, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_oeks15 = xr.open_dataset(file_path_oeks15)
        return file_path_oeks15, dataset_oeks15

    if type == 'spartacus':
        # Call the load_spartacus function to get the file path
        file_path_spartacus = load_spartacus(variable, start, end, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_spartacus = xr.open_dataset(file_path_spartacus)
        return file_path_spartacus, dataset_spartacus

    if type == 'eurocordex':
        # Call the load_cordex function to get the file path
        file_path_cordex = load_cordex(model_global, model_regional, variable, experiment, ens, start, end, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_cordex = xr.open_dataset(file_path_cordex)
        return file_path_cordex, dataset_cordex

    if type == 'eobs':
        # Call the load_eobs function to get the file path
        file_path_eobs = load_eobs(variable, start, end, output_path)

        # Use xarray to load the dataset from the downloadeed netCDF file
        dataset_eobs = xr.open_dataset(file_path_eobs)
        return file_path_eobs, dataset_eobs

    if type == 'destine':
        # Call the download function to get the file path
        file_path_destine = load_destine(model_global, variable, experiment, start, end, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_destine = xr.open_dataset(file_path_destine)
        return file_path_destine, dataset_destine

    if type == 'era5':
        # Call the download function to get the file path
        file_path_era5 = load_era5(model_global, variable, start, end, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_era5 = xr.open_dataset(file_path_era5)
        return file_path_era5, dataset_era5

    if type == 'ch2025':
        # Call the load_ch2025 function to get the file path
        file_path_ch2025 = load_ch2025(model_global, model_regional, variable, experiment, output_path)

        # Use xarray to load the dataset from the downloaded netCDF file
        dataset_ch2025 = xr.open_dataset(file_path_ch2025)
        return file_path_ch2025, dataset_ch2025

    else:
        raise ValueError(f"Unsupported type: {type}")


def regrid(type, target_file, input_file, output_path_regrid):
    """
    Function that calls the corresponding regridding function.
    Both the path of the file and the xarray object are returned.
    """

    if type == 'nneighbor':
        # nearest neighbor interpolation
        regrid_path = remapnn(target_file, input_file, output_path_regrid)

        dataset_regrid = xr.open_dataset(regrid_path)
        return regrid_path, dataset_regrid

    if type == 'bilinear':
        # bilinear interpolation
        regrid_path = remapbil(target_file, input_file, output_path_regrid)

        dataset_regrid = xr.open_dataset(regrid_path)
        return regrid_path, dataset_regrid

    if type == 'conservative':
        # conservative interpolation
        regrid_path = remapcon(target_file, input_file, output_path_regrid)

        dataset_regrid = xr.open_dataset(regrid_path)
        return regrid_path, dataset_regrid

    if type == 'distance':
        # distance weighted interpolation
        regrid_path = remapdis(target_file, input_file, output_path_regrid)

        dataset_regrid = xr.open_dataset(regrid_path)
        return regrid_path, dataset_regrid


def mask(target_grid, input_grid, output_path_mask):
    """
    Function that calls the masking function.
    Both the path of the file and the xarray object are returned.
    """
    mask_path = masking(target_grid, input_grid, output_path_mask)

    dataset_mask = xr.open_dataset(mask_path)
    return mask_path, dataset_mask



# era5_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import cdsapi
import shutil
import os
import xarray as xr
import numpy as np
from datetime import datetime, timedelta
from .dictionary import dictionary


def load_era5(model_global, variable, start, end, output_path):
    """
    Download ERA5 data from the Copernicus Climate Data Store using the CDS API.

        Input:
            model_global          str          name of the dataset
            variable_long         str          variable name
            start                 str          start date (e.g. 20200901)
            end                   str          end date (e.g. 20200930)
            output_path           str          path where netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the ERA5 data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    era5_info = dictionary[variable]['era5']
    variable_long = era5_info['name']
    standard_unit = dictionary[variable]['oeks15']['units']

    # Define filename
    file = f"{variable}_{model_global}_{start}-{end}.nc"
    output_file = os.path.join(output_path, file)

    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded ERA5 data successfully.")
        return output_file

    else:
        # Create temporary directory
        temp_dir = os.path.join(output_path, "tmp")
        os.makedirs(temp_dir, exist_ok=True)
        temp_file = os.path.join(temp_dir, file)

        # Download ERA5 data
        try:
            # Initialize the API client
            client = cdsapi.Client()

            # Convert strings to datetime objects
            start_date = datetime.strptime(start, "%Y%m%d")
            end_date = datetime.strptime(end, "%Y%m%d")

            # Generate list of years, months, and days
            date_range = [start_date + timedelta(days=x) for x in range((end_date - start_date).days + 1)]

            years = sorted(list(set(date.strftime("%Y") for date in date_range)))
            months = sorted(list(set(date.strftime("%m") for date in date_range)))
            days = sorted(list(set(date.strftime("%d") for date in date_range)))

            dataset = model_global
            request = {
                "product_type": ["reanalysis"],
                "variable": variable_long,
                "year": years,
                "month": months,
                "day": days,
                "area": [80, -50, 20, 75],
                "data_format": "netcdf",
            }

            # Daily statistic products
            if model_global in [
                "derived-era5-single-levels-daily-statistics",
                "derived-era5-land-daily-statistics",
            ]:

                # Use daily mean for temperature
                if variable == 'tas':
                    request.update({
                        "daily_statistic": "daily_mean",
                        "time_zone": "utc+00:00",
                        "frequency": "1_hourly",
                    })

                # Use daily_sum for precipitation, daily_mean for temperature
                if variable == 'pr' and model_global == "derived-era5-single-levels-daily-statistics":
                    request.update({
                        "daily_statistic": "daily_sum",
                        "time_zone": "utc+00:00",
                        "frequency": "1_hourly",
                    })

            # ERA5-Land total precipitation
            elif variable == 'pr' and model_global == 'reanalysis-era5-land':
                request["time"] = ["00:00"]
                request["download_format"] = ["unarchived"]

            client.retrieve(dataset, request, temp_file)
            print("Downloaded ERA5 data successfully.")

        except Exception as e:
            print(f"Failed to download ERA5 data.\nError: {e}")
            return None

        # Open previously downloaded dataset
        dataset = xr.open_dataset(temp_file)

        # Standardize ERA5 and ERA5-Land temperature
        if variable == 'tas':

            # Rename the variable and time coordinate
            dataset = dataset.rename({'t2m': variable, 'valid_time': 'time'})

            # Convert from kelvin to celsius
            dataset[variable] = dataset[variable] - 273.15
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

            # Write dataset to netCDF file
            dataset.to_netcdf(output_file)

        # Standardize ERA5 precipitation
        if variable == 'pr' and model_global == 'derived-era5-single-levels-daily-statistics':

            # Rename the variable and time coordiante
            dataset = dataset.rename({'tp': variable,  'valid_time': 'time'})

            # Convert from m to mm
            dataset[variable] = dataset[variable]*1000
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

            # Write dataset to netCDF file
            dataset.to_netcdf(output_file)

        # Standardize ERA5-Land precipitation
        if variable == 'pr' and model_global == 'reanalysis-era5-land':

            # Rename the variable and time coordinate
            dataset = dataset.rename({'tp': variable, 'valid_time': 'time'})

            # Assign the precipitation to the previous day explicitly:
            dataset = dataset.assign_coords(
                time=(dataset.time - np.timedelta64(1, 'D'))
            )

            # Convert from meters to millimeters
            dataset[variable] = dataset[variable]*1000
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

            # Write dataset to netCDF file
            dataset.to_netcdf(output_file)

        # Clean temporary directory
        shutil.rmtree(temp_dir)

        return output_file



import json
import sys
from getpass import getpass, getuser
from pathlib import Path
from typing import Annotated, Optional
from urllib.parse import parse_qs, urlparse

import requests
from conflator import CLIArg, ConfigModel, Conflator, EnvVar
from lxml import html
from pydantic import Field

IAM_URL = "https://auth.destine.eu"
CLIENT_ID = "polytope-api-public"
REALM = "desp"
SERVICE_URL = "https://polytope.lumi.apps.dte.destination-earth.eu/"


class Config(ConfigModel):
    user: Annotated[
        Optional[str],
        Field(description="Your DESP username"),
        CLIArg("-u", "--user"),
        EnvVar("USER"),
    ] = None
    password: Annotated[
        Optional[str],
        Field(description="Your DESP password"),
        CLIArg("-p", "--password"),
        EnvVar("PASSWORD"),
    ] = None
    outpath: Annotated[
        str,
        Field(description='The file to write the token to (or "stdout")'),
        CLIArg("-o", "--outpath"),
    ] = str(Path().home() / ".polytopeapirc")


config = Conflator("despauth", Config).load()

if config.user == None:
    config.user = input("Username: ")
if config.password == None:
    config.password = getpass(prompt="Password: ")

with requests.Session() as s:
    # Get the auth url
    auth_url = (
        html.fromstring(
            s.get(
                url=IAM_URL + "/realms/" + REALM + "/protocol/openid-connect/auth",
                params={
                    "client_id": CLIENT_ID,
                    "redirect_uri": SERVICE_URL,
                    "scope": "openid offline_access",
                    "response_type": "code",
                },
            ).content.decode()
        )
        .forms[0]
        .action
    )

    # Login and get auth code
    login = s.post(
        auth_url,
        data={
            "username": config.user,
            "password": config.password,
        },
        allow_redirects=False,
    )

    # We expect a 302, a 200 means we got sent back to the login page and there's probably an error message
    if login.status_code == 200:
        tree = html.fromstring(login.content)
        error_message_element = tree.xpath('//span[@id="input-error"]/text()')
        error_message = (
            error_message_element[0].strip()
            if error_message_element
            else "Error message not found"
        )
        raise Exception(error_message)

    if login.status_code != 302:
        raise Exception("Login failed")

    auth_code = parse_qs(urlparse(login.headers["Location"]).query)["code"][0]

    # Use the auth code to get the token
    response = requests.post(
        IAM_URL + "/realms/" + REALM + "/protocol/openid-connect/token",
        data={
            "client_id": CLIENT_ID,
            "redirect_uri": SERVICE_URL,
            "code": auth_code,
            "grant_type": "authorization_code",
            "scope": "",
        },
    )

    if response.status_code != 200:
        raise Exception("Failed to get token")

    # instead of storing the access token, we store the offline_access (kind of "refresh") token
    token = response.json()["refresh_token"]
    # offline_token = response.json()['refresh_token']

    if config.outpath != "stdout":
        with open(config.outpath, "w") as file:
            dico = {"user_key": token}
            json.dump(dico, file)
            print(f"Token successfully written to {config.outpath}")
    else:
        print(token)



# cordex_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import os
import wget
import glob
import shutil
import xarray as xr
from pyesgf.search import SearchConnection
from datetime import datetime
from cdo import Cdo
from .dictionary import dictionary


# Function to find matching URLs
def find_matching_urls(urls, target_start_dt, target_end_dt):
    matching_urls = []
    for url in urls:
        # Extract date range from URL
        parts = url.split("_")[-1].replace(".nc", "").split("-")
        start_dt = datetime.strptime(parts[0], "%Y%m%d")
        end_dt = datetime.strptime(parts[1], "%Y%m%d")

        # Check if the file falls within the target range
        if start_dt >= target_start_dt and end_dt <= target_end_dt:
            matching_urls.append(url)

    return matching_urls


def load_cordex(model_global, model_regional, variable, experiment, ens,
                start, end, output_path):
    """
    Download CORDEX data from ESGF using the ESGF Pyclient.

        Input:
            model_global       str      name of the global climate model
            model_regional     str      name of the regional climate model
            variable           str      variable name (e.g. tas)
            experiment         str      name of the experiment (eg. historical)
            start              str      start date (e.g. 20160101)
            end                str      end date (e.g. 20201231)
            output_directory   str      path where the netCDF file is stored

        Output:
            Returns the path of the netCDF file containing the CORDEX data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    eurocordex_info = dictionary[variable]['eurocordex']
    variable_long = eurocordex_info['name']
    standard_key = next(iter(dictionary[variable]))
    standard_unit = dictionary[variable][standard_key]['units']

    # Define output filename
    file = (
        f"{variable}_{model_global}_{experiment}_{ens}_"
        f"{model_regional}_{start}-{end}.nc"
    )
    output_file = os.path.join(output_path, file)

    # Check whether the file already exits and return the path of the file
    if os.path.exists(output_file):
        print("Loaded EURO-CORDEX data successfully.")
        return output_file

    else:
        # Create temporary directory
        temp_dir = os.path.join(output_path, "tmp")
        os.makedirs(temp_dir, exist_ok=True)
        temp_file = os.path.join(temp_dir, file)

        esgf_nodes = [
            "esgf-data.dkrz.de",
            "esg-dn1.nsc.liu.se",
#            "esgf-node.llnl.gov",
            "esgf.ceda.ac.uk",
        ]

        # Download EURO-CORDEX data
        for hostname in esgf_nodes:
            print(f"\nTrying node: {hostname}")
            try:
                url = f"https://{hostname}/esg-search"
                conn = SearchConnection(url, distrib=True)

                ctx = conn.new_context(
                    project="CORDEX",
                    domain="EUR-11",
                    driving_model=model_global,
                    experiment=experiment,
                    ensemble=ens,
                    rcm_name=model_regional,
                    variable=variable_long,
                    time_frequency="day",
                    latest=True
                )

                results = ctx.search()

                if not results:
                    print("No datasets found on this node.")
                    continue

                ds = results[0]
                files = ds.file_context().search()

                # Get list of URLs
                url_list = []
                for f in files:
                    url_list.append(f.download_url)

                # Convert target dates to datetime objects
                target_start_dt = datetime.strptime(start, "%Y%m%d")
                target_end_dt = datetime.strptime(end, "%Y%m%d")

                # Find matching URLs
                matching_urls = find_matching_urls(url_list, target_start_dt, target_end_dt)

                for url in matching_urls:
                    filename = os.path.join(temp_dir, os.path.basename(url))
                    wget.download(url, filename)
                    print("Downloaded EURO-CORDEX data successfully.")

            except Exception as e:
                print(f"Failed to download EURO-CORDEX data.\nError: {e}")
                return None

        # Use CDO to merge netcdf files
        daily_files = sorted(glob.glob(os.path.join(temp_dir, "*.nc")))

        # If cdo is not in the path, add it manually
        conda_bin = os.path.expanduser('~/.conda/envs/climxtract/bin')
        os.environ['PATH'] += f':{conda_bin}'

        # Initialize the Cdo object
        cdo = Cdo()

        if daily_files:
            # Convert the list of files into a space-separated string
            file_list = " ".join(daily_files)

            # Perform merging of files on the time dimension
            cdo.mergetime(input=file_list, output=temp_file)

        # Open previously downloaded dataset
        dataset = xr.open_dataset(temp_file)

        # Standardize variable and units for temperature
        if variable == 'tas':
            # Convert from Kelvin to Celsius
            dataset[variable] = dataset[variable] - 273.15
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

        # Standardize variable and units for precipitation
        if variable == 'pr':
            # Convert kg m-2 s-1 into kg m-2
            dataset[variable] = dataset[variable]*86400
            units = dataset[variable].attrs.get("units", None)
            if units is None or units not in standard_unit:
                print(f"Warning: {variable} has no or wrong unit attribute.")
                dataset[variable].attrs['units'] = standard_unit
                print(f"Note: Unit attribute of {variable} changed to {standard_unit}.")

        # Save dataset to netCDF file
        dataset.to_netcdf(output_file)

        # Clean temporary directory
        shutil.rmtree(temp_dir)

        return output_file


# ch2025_download.py
# !/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Maximilian Meindl, University of Vienna
"""

import re
import os
import xarray as xr
import urllib.request
from pystac_client import Client
from .dictionary import dictionary


def find_matching_urls(variable, gwl, model_global, model_regional):
    """
    """
    # Convert GWL numeric to label
    gwl_label = None
    if isinstance(gwl, (int, float)):
        gwl_label = f"gwl{gwl:.1f}".replace(".", "_").replace("_", ".")
    elif isinstance(gwl, str):
        gwl_label = gwl.lower()

    # Strict variable matching pattern
    variable = variable.lower()
    variable_pattern = re.compile(rf"_({variable})_", re.IGNORECASE)

    # Open CH2025 catalog
    catalog = Client.open("https://data.geo.admin.ch/api/stac/v1/")
    collection = catalog.get_collection(
        "ch.meteoschweiz.ogd-climate-scenarios-ch2025-grid"
    )

    # Collect all assets
    all_assets = {}
    for item in collection.get_items():
        all_assets |= item.assets

    # Filter matching assets
    matches = []

    for key, asset in all_assets.items():

        key_low = key.lower()

        # Strict variable match
        if not variable_pattern.search(key_low):
            continue

        if variable == "tas" and ("tasmin" in key_low or "tasmax" in key_low):
            continue

        # Filter GWL or ref
        if gwl_label:
            if gwl_label not in key_low:
                continue

        # Extract model string: "..._tas_smhi-rca-mpiesm_gwl3.0.nc"
        m = re.search(r"_(.*?)_(gwl|ref)", key_low)
        if not m:
            continue

        model_str = m.group(1)
        parts = model_str.split("-")

        if len(parts) < 3:
            continue

        inst = parts[0]
        rcm = parts[1]
        gcm = "-".join(parts[2:])

        # Filter by model_global
        if model_global:
            if model_global.lower() not in gcm:
                continue

        # Filter by model_regional
        if model_regional:
            if model_regional.lower() not in f"{inst}-{rcm}":
                continue

        matches.append((key, asset.href))

    if len(matches) == 0:
        print("No matching CH2025 data found for your criteria.")
        return None, None, None

    asset_key, href = matches[0]

    return asset_key, href


def load_ch2025(model_global, model_regional, variable, experiment, output_path):
    """
    Download CH2025 daily gridded dataset filtered by variable, global warming level
    
        Input:
            model_global         string      name of the global climate model
            model_regional       string      name of the regional model
            variable             string      variable name (tass, tasmin, tasmax, pr)
            experiment           string      global warming level (1.5, 2.0, 2.5, 3.0, ref91-20)
            ens                  string      name of ensemble member
            output_path          string      path where netCDF file is stored
            
        Output:
            Returns the path of the netCDF file containing the CH2025 data.
    """
    # Validate variable
    if variable not in dictionary:
        raise ValueError(f"Variable must be one of {list(dictionary.keys())}")

    ch2025_info = dictionary[variable]['ch2025']
    variable_long = ch2025_info['name']
    #standard_key = next(iter(dictionary[variable]))
    #standard_unit = dictionary[variable][standard_key]['units']
    
    # Define filename
    file = (
        f"{variable}_{model_global}_{experiment}_"
        f"{model_regional}.nc"
    )

    output_file = os.path.join(output_path, file)
    
    # Check whether the file already exists and return the path of the file
    if os.path.exists(output_file):
        print("Loaded CH2025 data successfully.")
        return output_file

    else:
        # Download CH2025 data
        try:
            #
            asset_key, href = find_matching_urls(variable, experiment, model_global, model_regional)
            print("Downloading:", asset_key)
            
            # Download the file using urllib
            urllib.request.urlretrieve(href, output_file)
            print("Downloaded CH2025 data successfully.")
            
        except Exception as e:
            print(f"Failed to download CH2025 data.\nError: {e}")

        return output_file